@@ -90,6 +90,7 @@ CREATE TABLE cmplx_groups (
   id int(10) NOT NULL auto_increment,
   group_name char(32) default '',
   description char(32) default '',
+  parentid int(10) default NULL,
   PRIMARY KEY (id)
 );
 
@@ -132,3 +133,18 @@ CREATE TABLE cmplx_groups_divisions_link (
   dvid int(10),
   PRIMARY KEY (id)
 );
+
+CREATE TABLE rcsv_groups (
+  id int(10) NOT NULL auto_increment,
+  group_name char(32) default '',
+  description char(32) default '',
+  parentid int(10) default NULL,
+  PRIMARY KEY (id)
+);
+
+CREATE TABLE rcsv_users_groups_link (
+  id int(10) NOT NULL auto_increment,
+  uid int(10),
+  gid int(10),
+  PRIMARY KEY (id)
+);
