@@ -3,11 +3,15 @@ package LISM::Storage::LDAP;
 use strict;
 use base qw(LISM::Storage);
 use Net::LDAP;
+use Net::LDAP::Control::Paged;
+use Net::LDAP::Control::VLV;
+use Net::LDAP::Control::Sort;
 use Net::LDAP::Constant qw(:all);
 use MIME::Base64;
 use Encode;
 use Data::Dumper;
 
+our $controlAttr = 'lismControl';
 our $rawattrs = '^(jpegphoto|photo|objectSid|objectGUID|.*;binary)$';
 
 =head1 NAME
@@ -300,6 +304,18 @@ sub modify
             last;
         }
 
+        if ($key =~ /^lismparentdn$/i) {
+            if ($action eq "REPLACE" && @values && $values[0]) {
+                if ($dn =~ /^[^,]+,$values[0]$/i) {
+                    return LDAP_SUCCESS;
+                } else {
+                    return $self->move($dn, $values[0]);
+                }
+            } else {
+                return LDAP_UNWILLING_TO_PERFORM;
+            }
+        }
+
         # Attribute Mapping
         foreach my $ldapmap (@{$conf->{ldapmap}}) {
             if ($ldapmap->{type} =~ /^objectclass$/i) {
@@ -388,7 +404,15 @@ sub add
     my %attrs;
     my @info = split(/\n/, $entryStr);
     foreach my $attr (@info) {
-        my ($key, $val) = split(/: /, $attr);
+        my $key;
+        my $val;
+        if ($attr =~ /:: /) {
+            ($key, $val) = split(/:: /, $attr);
+            $val = decode_base64($val);
+        } else {
+            ($key, $val) = split(/: /, $attr);
+        }
+
         if ($key =~ /^(createTimestamp|modifyTimestamp|plainpassword)$/i) {
             next;
         }
@@ -437,7 +461,6 @@ sub modrdn
 {
     my $self = shift;
     my ($dn, $newrdn, $delFlag) = @_;
-
     my $conf = $self->{_config};
 
     if ($self->_getConnect()) {
@@ -534,6 +557,54 @@ sub delete
 
 =pod
 
+=head2 move($dn, $parentdn)
+
+move information in LDAP directory.
+
+=cut
+
+sub move
+{
+    my $self = shift;
+    my ($dn, $parentdn) = @_;
+    my $conf = $self->{_config};
+
+    # DN mapping
+    foreach my $ldapmap (@{$conf->{ldapmap}}) {
+        if ($ldapmap->{type} =~ /^dn$/i) {
+            if ($dn =~ /^$ldapmap->{local}=/i && (!defined($ldapmap->{dn}) || $dn =~ /$ldapmap->{dn}/i)) {
+                $dn = $self->_rewriteDn($ldapmap, 'request', $dn);
+            }
+        } elsif ($ldapmap->{type} =~ /^attribute$/i) {
+            $dn =~ s/^$ldapmap->{local}=/$ldapmap->{foreign}=/i;
+        }
+    }
+
+    $dn =~ s/$self->{suffix}$/$conf->{nc}/i;
+    $parentdn =~ s/$self->{suffix}$/$conf->{nc}/i;
+
+    my ($rc, $entryStr) = $self->_do_search(undef, $dn, 0, 0, 1, 0, '(objectClass=*)');
+    if ($rc) {
+        return $rc;
+    }
+
+    my ($rdn) = ($dn =~ /^([^,]+),/);
+    $entryStr =~ s/^dn:.*\n//;
+    $rc = $self->add("$rdn,$parentdn", $entryStr);
+    if ($rc) {
+        return $rc;
+    }
+
+    $rc = $self->delete($dn);
+    if ($rc) {
+        return $rc;
+    }
+
+    return $rc;
+}
+
+=pod
+
 =head2 hashPasswd($passwd, $salt)
 
 add hash schema at the head of hashed password.
@@ -607,13 +678,13 @@ sub _freeConnect
 {
     my $self =shift;
     my $conf = $self->{_config};
-    my ($msg) = @_;
+    my ($msg, $free) = @_;
 
     if (!$msg) {
         return -1;
     }
 
-    if ($msg->code == LDAP_SERVER_DOWN || $msg->code == -1 ||
+    if ($msg->code == LDAP_SERVER_DOWN || $msg->code == -1 || $free ||
         (defined($conf->{connection}[0]->{type}) && $conf->{connection}[0]->{type}[0] eq 'every')) {
         $self->{ldap}->unbind();
         $self->{bind}->unbind();
@@ -691,8 +762,57 @@ sub _do_search
     my($control, $base, $scope, $deref, $sizeLim, $timeLim, $filterStr, $attrOnly, @attrs) = @_;
     my $conf = $self->{_config};
     my @match_entries = ();
+    my %ctrl_class;
+    my $pagenum = 0;
     my $rc = LDAP_SUCCESS;
 
+    if ($control) {
+        for (my $i = 0; $i < @{$control}; $i++) {
+            $ctrl_class{ref(${$control}[$i])} = $i;
+        }
+    }
+
+    my ($lismctrls) = ($filterStr =~ /^\(&\($controlAttr=([^\)]+)\)/i);
+    if ($lismctrls) {
+        $filterStr =~ s/^\(&\($controlAttr=[^\)]+\)//;
+        $filterStr =~ s/\)$//;
+        foreach my $lismctrl (split(/&/, $lismctrls)) {
+            my $ldapctrl;
+            my $class;
+            my ($key, $value) = split(/=/, $lismctrl);
+            if ($key eq 'paged') {
+                my $size;
+                ($size, $pagenum) = split(/,/, $value);
+                $ldapctrl = Net::LDAP::Control::Paged->new(size => $size);
+                $class = 'Net::LDAP::Control::Paged';
+            } elsif ($key eq 'vlv') {
+                my ($size, $offset) = split(/,/, $value);
+                $ldapctrl = Net::LDAP::Control::VLV->new(
+                    after => $size -1,
+                    before => 0,
+                    content => 0,
+                    offset => $offset
+                );
+                $class = 'Net::LDAP::Control::VLV';
+            } elsif ($key eq 'sort') {
+                $ldapctrl = Net::LDAP::Control::Sort->new(order => $value);
+                $class = 'Net::LDAP::Control::Sort';
+            } else {
+                next;
+            }
+            if (defined($ctrl_class{$class})) {
+                ${$control}[$ctrl_class{$class}] = $ldapctrl;
+            } else {
+                if ($control) {
+                    push(@{$control}, $ldapctrl);
+                } else {
+                    $control = [$ldapctrl];
+                }
+                $ctrl_class{$class} = @{$control} - 1;
+            }
+        }
+    }
+
     my $filter = Net::LDAP::Filter->new($filterStr);
     if (!defined($filter)) {
         return (LDAP_FILTER_ERROR, ());
@@ -734,10 +854,10 @@ sub _do_search
     $filterStr =~ s/$self->{suffix}(\)*)/$conf->{nc}$1/gi;
     $filterStr = encode('utf8', $filterStr);
 
+    my @rldapmap = reverse(@{$conf->{ldapmap}});
     my $msg;
-    my $class = $control ? ref(${$control}[0]) : undef;
-    while (1) {
-        if ($class) {
+    for (my $i = 1;;$i++) {
+        if (keys %ctrl_class) {
             $msg = $self->{ldap}->search(base => $base, scope => $scope, deref => $deref, sizelimit => $sizeLim, timelimit => $timeLim, filter => $filterStr, control => $control);
         } else {
             $msg = $self->{ldap}->search(base => $base, scope => $scope, deref => $deref, sizelimit => $sizeLim, timelimit => $timeLim, filter => $filterStr);
@@ -747,77 +867,79 @@ sub _do_search
             last;
         }
 
-        for (my $i = 0; $i < $msg->count; $i++) {
-            my $entry = $msg->entry($i);
-            my $dn = decode('utf8', $entry->dn);
+        if (!$pagenum || $pagenum == $i) {
+            for (my $j = 0; $j < $msg->count; $j++) {
+                my $entry = $msg->entry($j);
+                my $dn = decode('utf8', $entry->dn);
 
-            $dn =~ s/$conf->{nc}$/$self->{suffix}/i;
-            my $entryStr = "dn: $dn\n";
-            if ($dn =~ /^$self->{suffix}$/i) {
-                next;
-            } else {
-                foreach my $attr ($entry->attributes) {
-                    foreach my $value ($entry->get_value($attr)) {
-                        if ($attr =~ /$rawattrs/i) {
-                            $value = encode_base64($value, '');
-                            $entryStr = $entryStr.$attr.":: $value\n";
-                        } else {
-                            if ($value =~ /\n/) {
+                $dn =~ s/$conf->{nc}$/$self->{suffix}/i;
+                my $entryStr = "dn: $dn\n";
+                if ($dn =~ /^$self->{suffix}$/i) {
+                    next;
+                } else {
+                    foreach my $attr ($entry->attributes) {
+                        foreach my $value ($entry->get_value($attr)) {
+                            if ($attr =~ /$rawattrs/i) {
                                 $value = encode_base64($value, '');
                                 $entryStr = $entryStr.$attr.":: $value\n";
                             } else {
-                                $value = decode('utf8', $value);
-                                $value =~ s/$conf->{nc}$/$self->{suffix}/i;
-                                $entryStr = $entryStr.$attr.": $value\n";
+                                if ($value =~ /\n/) {
+                                    $value = encode_base64($value, '');
+                                    $entryStr = $entryStr.$attr.":: $value\n";
+                                } else {
+                                    $value = decode('utf8', $value);
+                                    $value =~ s/$conf->{nc}$/$self->{suffix}/i;
+                                    $entryStr = $entryStr.$attr.": $value\n";
+                                }
                             }
                         }
                     }
                 }
-            }
 
-            if (!$self->_checkEntry($entryStr)) {
-                next;
-            }
-
-            # Attribute mapping
-            foreach my $ldapmap (@{$conf->{ldapmap}}) {
-                if ($ldapmap->{type} =~ /^objectclass$/i) {
-                    $entryStr =~ s/^objectClass: $ldapmap->{foreign}$/objectClass: $ldapmap->{local}/mi;
-                } elsif ($ldapmap->{type} =~ /^attribute$/i) {
-                    $entryStr =~ s/^$ldapmap->{foreign}:/$ldapmap->{local}:/gmi;
-                    $entryStr =~ s/: $ldapmap->{foreign}=/: $ldapmap->{local}=/gi;
+                if (!$self->_checkEntry($entryStr)) {
+                    next;
                 }
-            }
 
-            # DN mapping
-            foreach my $ldapmap (@{$conf->{ldapmap}}) {
-                if ($ldapmap->{type} =~ /^dn$/i) {
-                    my ($dn) = ($entryStr =~ /^dn: (.*)$/m);
-                    if ($dn =~ /^$ldapmap->{foreign}=/i && (!defined($ldapmap->{dn}) || $dn =~ /$ldapmap->{dn}/i)) {
-                        my ($rdn, $rdn_val) = ($entryStr =~ /^($ldapmap->{local}): (.*$)/mi);
-                        if ($rdn && $rdn_val) {
-                            $entryStr =~ s/^dn: [^,]+/dn: $rdn=$rdn_val/i;
+                # Attribute mapping
+                foreach my $ldapmap (@rldapmap) {
+                    if ($ldapmap->{type} =~ /^objectclass$/i) {
+                        $entryStr =~ s/^objectClass: $ldapmap->{foreign}$/objectClass: $ldapmap->{local}/mi;
+                    } elsif ($ldapmap->{type} =~ /^attribute$/i) {
+                        $entryStr =~ s/^$ldapmap->{foreign}:/$ldapmap->{local}:/gmi;
+                        $entryStr =~ s/: $ldapmap->{foreign}=/: $ldapmap->{local}=/gi;
+                    }
+                }
+
+                # DN mapping
+                foreach my $ldapmap (@rldapmap) {
+                    if ($ldapmap->{type} =~ /^dn$/i) {
+                        my ($dn) = ($entryStr =~ /^dn: (.*)$/m);
+                        if ($dn =~ /^$ldapmap->{foreign}=/i && (!defined($ldapmap->{dn}) || $dn =~ /$ldapmap->{dn}/i)) {
+                            my ($rdn, $rdn_val) = ($entryStr =~ /^($ldapmap->{local}): (.*$)/mi);
+                            if ($rdn && $rdn_val) {
+                                $entryStr =~ s/^dn: [^,]+/dn: $rdn=$rdn_val/i;
+                            }
                         }
                     }
                 }
-            }
 
-            push(@match_entries, $entryStr);
+                push(@match_entries, $entryStr);
+            }
         }
 
-        if ($class eq 'Net::LDAP::Control::Paged') {
+        if (defined($ctrl_class{'Net::LDAP::Control::Paged'})) {
             # Get cookie from paged control
-            my($resp) = $msg->control(LDAP_CONTROL_PAGED) or last;
+            my ($resp) = $msg->control(LDAP_CONTROL_PAGED) or last;
             my $cookie = $resp->cookie or last;
 
             # Set cookie in paged control
-            ${$control}[0]->cookie($cookie);
+            ${$control}[$ctrl_class{'Net::LDAP::Control::Paged'}]->cookie($cookie);
         } else {
             last;
         }
     }
 
-    $self->_freeConnect($msg);
+    $self->_freeConnect($msg, defined($ctrl_class{'Net::LDAP::Control::VLV'}));
 
     return ($msg->code , @match_entries);
 }
