@@ -140,7 +140,7 @@ sub _connect
     my $conf = $self->{_config};
 
     if (defined(${$db}) && (!defined($conf->{connection}[0]->{type}) || $conf->{connection}[0]->{type}[0] ne 'every')) {
-        my $sth = ${$db}->prepare("select 'check';");
+        my $sth = ${$db}->prepare("select 'check' from $conf->{connection}[0]->{table}");
         if (!$sth->execute) {
             $self->log(level => 'err', message => "Connection check failed");
             ${$db}->disconnect();
@@ -192,7 +192,7 @@ sub _checkConfig
         my $oconf = $conf->{object}{$oname};
 
         # check container fromtbls
-        if (defined($oconf->{container}[0]->{oname})) {
+        if (defined($oconf->{container}) && defined($oconf->{container}[0]->{oname})) {
             if (defined($oconf->{container}[0]->{fromtbls})) {
                 my $poconf = $self->{object}{$oconf->{container}[0]->{oname}[0]}->{conf};
                 if (",$oconf->{container}[0]->{fromtbls}[0]," =~
@@ -216,6 +216,10 @@ sub _checkConfig
                 }
             }
         }
+
+        if (!defined($conf->{connection}[0]->{table})) {
+            $conf->{connection}[0]->{table} = $oconf->{table}[0];
+        }
     }
 
     return 0;
@@ -235,11 +239,31 @@ sub _objSearch
     my ($obj, $pkeys, $suffix, $sizeLim, $filter, $attrOnly, @attrs) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
-    my $pkey = ${$pkeys}[$#{$pkeys}];
+    my $pkey = $self->_getPid($pkeys);
     my @match_entries = ();
     my @match_keys = ();
     my $rc = LDAP_SUCCESS;
 
+    # do plugin
+    if (defined($oconf->{plugin})) {
+        my $error;
+        my $keys;
+        my @entries;
+
+        ($rc, $keys, @entries) = $self->_doPlugin('search', $obj, $pkeys, $suffix, $sizeLim, $filter);
+        if ($rc) {
+            return ($rc, \@match_keys, @match_entries);
+        }
+
+        push(@match_keys, @{$keys});
+        push(@match_entries, @entries);
+
+    }
+
+    if (!defined($oconf->{id})) {
+        return ($rc, \@match_keys, @match_entries);
+    }
+
     DO: {
         # get data of the entries
         my $sql = "select distinct $oconf->{table}[0].$oconf->{id}[0]->{column}[0]";
@@ -260,12 +284,13 @@ sub _objSearch
         }
 
         # entries below suffix
-        if (defined($pkey) && defined($oconf->{container}) && !defined($oconf->{container}[0]->{rdn})) {
+        if (defined($pkeys) && defined($oconf->{container}) &&
+            defined($oconf->{container}[0]->{oname})) {
             my $poconf = $self->{object}{$oconf->{container}[0]->{oname}[0]}->{conf};
 
             if ($oconf->{table}[0] ne $poconf->{table}[0]) {
                 $from = "$from, $poconf->{table}[0]";
-                $where = "$where and $poconf->{table}[0].$poconf->{id}[0]->{column}[0] = \'$pkey\'";
+                $where = "$where and $poconf->{table}[0].$poconf->{id}[0]->{column}[0] ".(defined($pkey) ? "= \'$pkey\'" : "is NULL");
             }
 
             if (defined($oconf->{container}[0]->{fromtbls})) {
@@ -274,6 +299,9 @@ sub _objSearch
 
             if (defined($oconf->{container}[0]->{joinwhere})) {
                 my $pwhere = $oconf->{container}[0]->{joinwhere}[0];
+                if (!defined($pkey)) {
+                    $pwhere =~ s/= '?%c'?/is NULL/;
+                }
                 $pwhere = $self->_containerParse($pwhere, @{$pkeys});
                 $where = "$where and $pwhere";
             }
@@ -335,7 +363,7 @@ sub _objSearch
                         my @values = split($oconf->{attr}{$attr}->{delim}[0], $data[$j]);
                         foreach my $value (@values) {
                             $value =~ s/ *$//;
-                            if ($value) {
+                            if (!defined($value) || $value ne '') {
                                 if ($attr =~ /$rawattrs/i) {
                                     $value = encode_base64($value, '');
                                     $entry = $entry.$attr.":: $value\n";
@@ -351,7 +379,7 @@ sub _objSearch
                     } else {
                         my $value = $data[$j];
                         $value =~ s/ *$//;
-                        if ($value) {
+                        if (!defined($value) || $value ne '') {
                             if ($attr =~ /$rawattrs/i) {
                                 $value = encode_base64($value, '');
                                 $entry = $entry.$attr.":: $value\n";
@@ -417,7 +445,7 @@ sub _objModify
     my ($obj, $pkeys, $dn, @list) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
-    my $pkey = ${$pkeys}[$#{$pkeys}];
+    my $pkey = $self->_getPid($pkeys);
     my $rc = LDAP_SUCCESS;
 
     DO: {
@@ -677,7 +705,7 @@ sub _objAdd
     my ($obj, $pkeys, $dn, $entryStr) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
-    my $pkey = ${$pkeys}[$#{$pkeys}];
+    my $pkey = $self->_getPid($pkeys);
     my $rc = LDAP_SUCCESS;
 
     # check whether the entry already exists
@@ -788,7 +816,7 @@ sub _objAdd
         my $filter = Net::LDAP::Filter->new("(".encode('utf8', $rdn).")");
         my $keys;
 
-        ($rc, $keys) = $self->_objSearch($obj, [], $pdn, -1, $filter);
+        ($rc, $keys) = $self->_objSearch($obj, undef, $pdn, -1, $filter);
         if ($rc || !@{$keys}) {
             $self->log(level => 'err', message => "Can't get id of $dn from the table");
             $rc = LDAP_OPERATIONS_ERROR;
@@ -811,6 +839,7 @@ sub _objAdd
 	        $sql = $addproc;
                 $sql =~ s/\%o/$key/g;
                 $sql =~ s/\%v/$si_values[$i]/g;
+                $sql =~ s/\%c/$pkey/g;
                 $sql = $self->_funcParse($sql, $dn, $entryStr);
 
                 # multibyte value
@@ -838,7 +867,8 @@ sub _objAdd
         }
 
         # add the link with container
-        if (defined($oconf->{container}) && defined($oconf->{container}[0]->{addproc})) {
+        if (defined($pkey) && defined($oconf->{container}) &&
+            defined($oconf->{container}[0]->{addproc})) {
             $sql = $oconf->{container}[0]->{addproc}[0];
             $sql =~ s/\%o/$key/g;
             $sql = $self->_containerParse($sql, @{$pkeys});
@@ -874,7 +904,7 @@ sub _objDelete
     my ($obj, $pkeys, $dn) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
-    my $pkey = ${$pkeys}[$#{$pkeys}];
+    my $pkey = $self->_getPid($pkeys);
     my $rc = LDAP_SUCCESS;
 
     DO: {
@@ -939,14 +969,15 @@ sub _objDelete
 
         if (!defined($oconf->{noop}) || !grep(/^delete$/i, @{$oconf->{noop}})) {
             # delete the link with container
-            if (defined($oconf->{container}) && defined($oconf->{container}[0]->{delproc})) {
+            if (defined($pkey) && defined($oconf->{container}) &&
+                defined($oconf->{container}[0]->{delproc})) {
                 $sql = $oconf->{container}[0]->{delproc}[0];
                 $sql =~ s/\%o/$key/g;
                 $sql = $self->_containerParse($sql, @{$pkeys});
                 $sql = $self->_funcParse($sql, $dn);
                 $sth = $self->{db}->prepare("$sql");
                 if (!$sth->execute) {
-                    $self->log(level => 'err', message => "Adding link of container by \"$sql\" failed: ".$sth->errstr);
+                    $self->log(level => 'err', message => "Deleting link of container by \"$sql\" failed: ".$sth->errstr);
                     $rc = LDAP_OPERATIONS_ERROR;
                     last DO;
                 }
@@ -979,12 +1010,101 @@ sub _objDelete
     return $rc;
 }
 
+=pod
+
+=head2 _objMove($obj, $pekys, $dn, $newpkeys, $newdn)
+
+Move the object.
+
+=cut
+
+sub _objMove
+{
+    my $self = shift;
+    my ($obj, $pkeys, $dn, $newpkeys, $newdn) = @_;
+    my $conf = $self->{_config};
+    my $oconf = $obj->{conf};
+    my $pkey = $self->_getPid($pkeys);
+    my $newpkey = $self->_getPid($newpkeys);
+    my $sql;
+    my $sth;
+    my $rc = LDAP_SUCCESS;
+
+    DO: {
+        # start transaction
+        if (!$self->{db}->begin_work) {
+            $self->log(level => 'err', message => "Can't begin: ".$self->{db}->errstr);
+            $rc = LDAP_OPERATIONS_ERROR;
+            last DO;
+        }
+
+        my $entry;
+        my $key;
+        my $newkey;
+
+        ($rc, $key, $entry) = $self->_baseSearch($obj, $pkeys, $dn, 0, 0, 1, 0, undef);
+        if ($rc) {
+            last DO;
+        }
+
+        if (!$entry) {
+            $rc = LDAP_NO_SUCH_OBJECT;
+            last DO;
+        }
+
+        ($rc, $newkey, $entry) = $self->_baseSearch($obj, $newpkeys, $newdn, 0, 0, 1, 0, undef);
+        if ($entry) {
+            $rc = LDAP_ALREADY_EXISTS;
+            last DO;
+        }
+
+        # add the link with container
+        if (defined($newpkey) && defined($oconf->{container}) &&
+            defined($oconf->{container}[0]->{addproc})) {
+            $sql = $oconf->{container}[0]->{addproc}[0];
+            $sql =~ s/\%o/$key/g;
+            $sql = $self->_containerParse($sql, @{$newpkeys});
+            $sql = $self->_funcParse($sql, $newdn, $entry);
+            $sth = $self->{db}->prepare("$sql");
+            if (!$sth->execute) {
+                $self->log(level => 'err', message => "Adding link of container by \"$sql\" failed: ".$sth->errstr);
+                $rc = LDAP_OPERATIONS_ERROR;
+                last DO;
+            }
+            $sth->finish;
+        }
+
+        # delete the link with container
+        if (defined($pkey) && defined($oconf->{container}) &&
+            defined($oconf->{container}[0]->{delproc})) {
+            $sql = $oconf->{container}[0]->{delproc}[0];
+            $sql =~ s/\%o/$key/g;
+            $sql = $self->_containerParse($sql, @{$pkeys});
+            $sql = $self->_funcParse($sql, $dn);
+            $sth = $self->{db}->prepare("$sql");
+            if (!$sth->execute) {
+                $self->log(level => 'err', message => "Deleting link of container by \"$sql\" failed: ".$sth->errstr);
+                $rc = LDAP_OPERATIONS_ERROR;
+                last DO;
+            }
+            $sth->finish;
+        }
+    }
+
+    if ($rc) {
+        $self->rollback();
+    }
+
+    return $rc;
+}
+
 sub _getParentRdn
 {
     my $self = shift;
-    my ($obj, $key) = @_;
+    my ($obj, $key, $pobj) = @_;
     my $conf = $self->{_config};
     my $oconf = $obj->{conf};
+    my $poconf = $pobj->{conf};
     my $selexpr;
     my $from;
     my $where;
@@ -996,8 +1116,6 @@ sub _getParentRdn
         return undef;
     }
 
-    my $poconf = $self->{object}{$obj->{parent}}->{conf};
-
     if (defined($poconf->{entry})) {
         return $poconf->{entry}[0]->{rdn}[0];
     }
@@ -1015,7 +1133,7 @@ sub _getParentRdn
     $where = "$oconf->{table}[0].$oconf->{id}[0]->{column}[0] = \'$key\'";
     if (defined($oconf->{container}[0]->{joinwhere})) {
         if ($oconf->{container}[0]->{joinwhere}[0] =~ /\%c/) {
-            my ($pidcol) = ($oconf->{container}[0]->{joinwhere}[0] =~ /([^ ]+) *= *\%c/);
+            my ($pidcol) = ($oconf->{container}[0]->{joinwhere}[0] =~ /([^ ]+) *= *'?\%c'?/);
             $where = "$poconf->{table}[0].$poconf->{id}[0]->{column}[0] = (select $pidcol from $oconf->{table}[0] where $where)";
         } else {
             $where = "$where and $oconf->{container}[0]->{joinwhere}[0]";
@@ -1036,7 +1154,11 @@ sub _getParentRdn
     my $rdn_val = $data[1];
     $rdn_val =~ s/ *$//;
 
-    return ("$poconf->{rdn}[0]=$rdn_val", $data[0]);
+    if ($data[0]) {
+        return ("$poconf->{rdn}[0]=$rdn_val", $data[0]);
+    } else {
+        return '';
+    }
 }
 
 sub _getAttrValues
@@ -1083,6 +1205,8 @@ sub _getAttrValues
         }
     } elsif (defined($oconf->{attr}{$attr}->{selexpr})){
         $selexpr = $oconf->{attr}{$attr}->{selexpr}[0];
+    } elsif (defined($oconf->{attr}{$attr}->{constant})) {
+        return "$attr: $oconf->{attr}{$attr}->{constant}[0]\n";
     } else {
         return '';
     }
@@ -1103,6 +1227,14 @@ sub _getAttrValues
         }
     }
 
+    if ($aoconf) {
+        foreach my $strginfo (@{$aoconf->{strginfo}}) {
+            if (defined($strginfo->{selwhere})) {
+                $where = "$where and $strginfo->{selwhere}[0]";
+            }
+        }
+    }
+
     my $sql = "select $selexpr from $from where $where";
     my $sth = $self->{db}->prepare($sql);
     if (!$sth->execute) {
