@@ -419,7 +419,7 @@ sub compare_regex {
   my ($file1, $file2) = @_;
 
   return File::Compare::compare_text($file1, $file2,
-      sub {$_[0] !~ /$_[1]/});
+      sub {chop($_[1]);$_[0] !~ /$_[1]/});
 }
 
 sub compare_log {
